/*
 *   Heat conduction demo program
 *
 *  solves the heat equation on a 2D grid
 *  and creates output for gnuplot.
 *
 *  March 2009
 *  Matthias.Lieber@tu-dresden.de
 *  Tobias.Hilbrich@zih.tu-dresden.de
 *
 *  Serial version.
 */

#include "exampleC.h"
#include "papi_wrap.h"

/******************************************************
 * Allocate the heatGrid and initialize all variables.
 ******************************************************/
void heatAllocate(heatGrid *grid, int xsize, int ysize)
{
    int i,j;

    grid->xsize = xsize;
    grid->ysize = ysize;
    grid->theta    = (double**) malloc (sizeof(double*)*(xsize+2));
    grid->thetanew = (double**) malloc (sizeof(double*)*(xsize+2));
    grid->theta   [0] = (double*) malloc (sizeof(double)*(xsize+2)*(ysize+2));
    grid->thetanew[0] = (double*) malloc (sizeof(double)*(xsize+2)*(ysize+2));

    for (i = 0; i < xsize+2; i++)
    {
        grid->theta   [i] = grid->theta   [0]+i*(ysize+2);
        grid->thetanew[i] = grid->thetanew[0]+i*(ysize+2);

        for (j = 0; j < ysize+2; j++)
        {
            grid->theta   [i][j] = 0.0;
            grid->thetanew[i][j] = 0.0;
        }
    }

    grid->dx = 1.0;
    grid->dy = 1.0;
    grid->k = 1.0;
}

/******************************************************
 * Deallocate the heatGrid.
 ******************************************************/
void heatDeallocate(heatGrid *grid)
{
    free (grid->theta[0]);
    free (grid->theta);
    grid->theta = NULL;
    free (grid->thetanew[0]);
    free (grid->thetanew);
    grid->thetanew = NULL;
}

/******************************************************
 * Initialize the grid with some meaninful start values.
 ******************************************************/
void heatInitialize(heatGrid* grid)
{
    double xcenter, ycenter, radius, cr;
    int x, y;

    /* initialize with a circle */
    for (x = 0; x < grid->xsize+2; x++)
        for (y = 0; y < grid->ysize+2; y++)
            grid->theta[x][y] = 0.0;

    xcenter = (grid->xsize - 10) * grid->dx / 2.0;
    ycenter = (grid->ysize + 6)  * grid->dy / 2.0;
    radius = fmin(grid->xsize * grid->dx, grid->ysize * grid->dy) * 0.25;

    for (x=1; x <= grid->xsize; x++)
    {
        for (y=1; y <= grid->ysize; y++)
        {
            cr = sqrt( (x-xcenter)*(x-xcenter) + (y-ycenter)*(y-ycenter) );
            if ( cr < radius )
            {
                /* sharp boundary */
                grid->theta[x][y] = grid->theta[x][y] + 2.0;

            }
        }
    }
}

/******************************************************
 * Print grid to console.
 ******************************************************/
void heatPrint(heatGrid* grid)
{
    int x, y;

    /* print header */
    for (x=1; x <= grid->xsize; x++)
        printf ("====");
    printf("\n");

    /* print data */
    for (y=1; y <= grid->ysize; y++)
    {
        for (x=1; x <= grid->xsize; x++)
        {
            if(grid->theta[x][y] < 1.e-100)
                printf ("  . ");
            else
                printf("%4.1f", grid->theta[x][y]);
        }
        printf("\n");
    }
}

/******************************************************
 * Calculate one timestep of size dt on the grid.
 *
 * for each grid point:
 *
 *                                  ( d2T   d2T )
 *          thetanew =  theta + k * ( --- + --- ) * dt
 *                                  ( dx2   dy2 )
 *                                  _____________
 *                                        |
 *                                        |
 *  dthetamax returns the max. value of this term
 *  (useful as exit condition for the time stepping)
 ******************************************************/
void heatTimestep(heatGrid* grid, double dt, double* dthetamax)
{
    int x, y;
    double dtheta;

    *dthetamax = 0.0;

    /* calculate the time step: read from theta, write new timestep to thetanew */
    for (x=1; x <= grid->xsize;x++)
    {
        for (y=1; y <= grid->ysize; y++)
        {
            dtheta = ( grid->theta[x-1][y] + grid->theta[x+1][y] - 2*grid->theta[x][y]) / (grid->dx * grid->dx)
                   + ( grid->theta[x][y-1] + grid->theta[x][y+1] - 2*grid->theta[x][y]) / (grid->dy * grid->dy);
            grid->thetanew[x][y] = grid->theta[x][y] + grid->k * dtheta * dt;

            *dthetamax = fmax(fabs(dtheta), *dthetamax); /* save max theta for the exit condition */
        }
    }

    /* update theta: copy thetanew to theta */
    for (x=1; x <= grid->xsize; x++)
        for (y=1; y <= grid->ysize; y++)
             grid->theta[x][y] = grid->thetanew[x][y];
}

/******************************************************
 * Set periodic boundary conditions.
 *
 * The grid arrays are allocated with additional "ghost cells"
 * in each spatial dimension. The lower boundary is copied to
 * the upper ghost cells (and vice versa) for each dimension:
 *
 *    ___________         ___________
 *   |  _______ |        |  stuvwx  |
 *   | |abcdef| |        |f|abcdef|a|
 *   | |ghijkl| |        |l|ghijkl|g|
 *   | |mnopqr| |   ->   |r|mnopqr|m|
 *   | |stuvwx| |        |x|stuvwx|s|
 *   |__________|        |__abcdef__|
 *
 ******************************************************/
void heatBoundary(heatGrid* grid)
{
    int i;

    /* y direction */
    for (i = 1; i <= grid->xsize; i++)
    {
        /* y-bottom */
        grid->theta[i][0] = grid->theta[i][grid->ysize];
        /* y-top */
        grid->theta[i][grid->ysize+1] = grid->theta[i][1];
    }

    /* x direction */
    for (i = 1; i <= grid->ysize; i++)
    {
        /* x-left */
        grid->theta[0][i] = grid->theta[grid->xsize][i];
        /* x-right */
        grid->theta[grid->xsize+1][i] = grid->theta[1][i];
    }
}

/******************************************************
 * Output results to gnuplot data file.
 *
 * Filename is: data-<filenumber>.dat
 ******************************************************/
void heatOutput (heatGrid* grid, int filenumber)
{
    char filename[64];
    int x, y;
    FILE *file;

    /* open file */
    sprintf(filename,"data-%08d.dat",filenumber);
    file = fopen (filename,"w");
    if(!file)
    {
        fprintf (file, "cannot open %s aborting\n",filename);
    }

    /* print data */
    for (y=1; y <= grid->ysize; y++)
        for (x=1; x <= grid->xsize; x++)
            fprintf (file, "%4d %4d %8.4f\n", x, y, grid->theta[x][y]);

    fclose (file);
}

/******************************************************
 * Calculate the total energy (sum of theta in all grid cells)
 ******************************************************/
void heatTotalEnergy(heatGrid* grid, double *energy)
{
  int x, y;

  *energy = 0.0;
  for (x=1; x <= grid->xsize; x++)
      for (y=1; y <= grid->ysize; y++)
          *energy += grid->theta[x][y];
}

/******************************************************
 * Main program and time stepping loop.
 ******************************************************/
int main (int argc, char** argv)
{
    heatGrid mygrid;
    double dt, dthetamax, energyInitial, energyFinal;
    int step, nsteps;
    int64_t t0,t1;
    long long v1[4], v0[4];
    char *sList[5];
    int i;

    /* create heatGrid and initialize variables */
#ifdef SMALL
    heatAllocate(&mygrid,   20,   20);
    nsteps = 1000;
#else
    heatAllocate(&mygrid, 2048, 2048);
    nsteps = 20;
#endif
    heatInitialize(&mygrid);
    dt = 0.05;
    dthetamax = 100.0;

#ifdef SMALL
    /* output of initial grid */
    printf("intial grid:\n");
    heatPrint(&mygrid);
    printf("\n");
#endif

    /* Init PAPI */
    sList[0]="PAPI_L1_DCM";
    sList[1]="PAPI_L2_DCM";
    sList[2]="PAPI_FP_INS";
    sList[3]="PAPI_RES_STL";
    sList[4]=NULL; 
    initPAPI(sList);

    /* energy of initial grid */
    heatTotalEnergy(&mygrid, &energyInitial);
    
    readPAPIcounters(v0);
    gettimec(&t0);

    /* time stepping loop */
    for( step=0 ; step<nsteps ; step++ )
    {
        heatBoundary(&mygrid);
        heatTimestep(&mygrid, dt, &dthetamax );
    }

    gettimec(&t1);
    readPAPIcounters(v1);
    stopPAPI();

#ifdef SMALL
    /* output of final grid */
    printf("final grid:\n");
    heatPrint(&mygrid);
    printf("\n");
#endif

    printf("iteration steps: %10d   dthetamax: %20.10f\n",nsteps, dthetamax);
    printf("%10.5f s\n",((float)(t1-t0))/1.0e6);

    printf("\n---- PAPI ------------------counters --- frequency ---------------\n");
    for(i=0;sList[i]!=NULL;i++)
    {
        printf("%-20s\t%12lld\t%f G/s\n", sList[i],v1[i]-v0[i], ((float)(v1[i]-v0[i]))/(t1-t0)/1000);
    }

    /* energy of final grid */
    heatTotalEnergy(&mygrid, &energyFinal);

    printf("\n= Energy Conservation Check =\n");
    printf(" initial Energy: %20.3f\n",energyInitial);
    printf("   final Energy: %20.3f\n",energyFinal);
    printf("     Difference: %20.3f\n",energyFinal-energyInitial);

    heatDeallocate(&mygrid);

    return 0;
}

/*EOF*/
